=== modified file 'bzrlib/revisionspec.py'
--- bzrlib/revisionspec.py	2007-12-03 21:11:03 +0000
+++ bzrlib/revisionspec.py	2007-12-03 21:35:54 +0000
@@ -396,14 +396,13 @@
         # representation.
         revision_id = osutils.safe_revision_id(self.spec, warn=False)
         if need_revno:
-            if revs is None:
-                # XXX Branch needs a method for incremental
-                # revision_id -> revno resolving without loading
-                # the whole history. (Lukas Lalinsky, 20081203)
-                revs = branch.revision_history()
-            return RevisionInfo.from_revision_id(branch, revision_id, revs)
+            try:
+                revno = branch.revision_id_to_revno(revision_id)
+            except errors.NoSuchRevision:
+                revno = None
         else:
-            return RevisionInfo(branch, None, revision_id)
+            revno = None
+        return RevisionInfo(branch, revno, revision_id)
 
     def in_branch(self, branch, need_revno=True):
         # Same as RevisionSpec.in_history, but without history loading.
@@ -481,8 +480,12 @@
 
     prefix = 'before:'
     
-    def _match_on(self, branch, revs):
-        r = RevisionSpec.from_string(self.spec)._match_on(branch, revs)
+    def _match_on(self, branch, revs, need_revno=True, in_branch=False):
+        revspec = RevisionSpec.from_string(self.spec)
+        if in_branch:
+            r = revspec.in_branch(branch, need_revno)
+        else:
+            r = revspec._match_on(branch, revs)
         if r.revno == 0:
             raise errors.InvalidRevisionSpec(self.user_spec, branch,
                                          'cannot go before the null: revision')
@@ -507,6 +510,9 @@
                                                  branch)
         return RevisionInfo(branch, revno, revision_id)
 
+    def in_branch(self, branch, need_revno=True):
+        return self._match_on(branch, None, need_revno, True)
+
 SPEC_TYPES.append(RevisionSpec_before)
 
 
@@ -520,11 +526,20 @@
 
     prefix = 'tag:'
 
-    def _match_on(self, branch, revs):
+    def _match_on(self, branch, revs, need_revno=True):
         # Can raise tags not supported, NoSuchTag, etc
-        return RevisionInfo.from_revision_id(branch,
-            branch.tags.lookup_tag(self.spec),
-            revs)
+        revision_id = branch.tags.lookup_tag(self.spec)
+        if need_revno:
+            try:
+                revno = branch.revision_id_to_revno(revision_id)
+            except errors.NoSuchRevision:
+                revno = None
+        else:
+            revno = None
+        return RevisionInfo(branch, revno, revision_id)
+
+    def in_branch(self, branch, need_revno=True):
+        return self._match_on(branch, None, need_revno)
 
 SPEC_TYPES.append(RevisionSpec_tag)
 

=== modified file 'bzrlib/status.py'
--- bzrlib/status.py	2007-12-03 20:48:08 +0000
+++ bzrlib/status.py	2007-12-03 21:35:54 +0000
@@ -84,13 +84,15 @@
             old = new.basis_tree()
         elif len(revision) > 0:
             try:
-                rev_id = revision[0].in_branch(wt.branch, False).rev_id
+                rev_id = revision[0].in_branch(wt.branch,
+                                               need_revno=False).rev_id
                 old = wt.branch.repository.revision_tree(rev_id)
             except errors.NoSuchRevision, e:
                 raise errors.BzrCommandError(str(e))
             if (len(revision) > 1) and (revision[1].spec is not None):
                 try:
-                    rev_id = revision[1].in_branch(wt.branch, False).rev_id
+                    rev_id = revision[1].in_branch(wt.branch,
+                                                   need_revno=False).rev_id
                     new = wt.branch.repository.revision_tree(rev_id)
                     new_is_working_tree = False
                 except errors.NoSuchRevision, e:

